Index: dod.c
===================================================================
RCS file: /cvs/public/parrot/dod.c,v
retrieving revision 1.13
diff -u -r1.13 dod.c
--- dod.c	17 Aug 2002 01:11:08 -0000	1.13
+++ dod.c	17 Aug 2002 07:57:02 -0000
@@ -62,7 +62,9 @@
     last = mark_used(current, last);
 
     /* Find important stuff on the system stack */
+/* not used in African Grey
     last = trace_system_stack(interpreter,last);
+*/
 
     /* Now, go run through the PMC registers and mark them as live */
     /* First mark the current set. */
@@ -275,7 +277,8 @@
         for (i = 0; i < cur_arena->used; i++) {
             /* If it's not live or on the free list, put it on the free list */
             if (!(pmc_array[i].flags & (PMC_live_FLAG | PMC_on_free_list_FLAG |
-                                       PMC_constant_FLAG))) {
+                                       PMC_constant_FLAG))
+                && pmc_array[i].last_seen != interpreter->cycle) {
                 add_free_pmc(interpreter,
                                 interpreter->arena_base->pmc_pool,
                                 &pmc_array[i]);
@@ -311,10 +314,16 @@
             /* If it's not live or on the free list, put it on the free list */
             if (!(b->flags & (BUFFER_live_FLAG | BUFFER_on_free_list_FLAG)) &&
                 (!(b->flags & BUFFER_constant_FLAG) || 
-                 (b->flags & BUFFER_COW_FLAG))) 
+                 (b->flags & BUFFER_COW_FLAG))
+                && b->last_seen != interpreter->cycle) 
             {
                 if (pool->mem_pool) {
                     ((struct Memory_Pool *)pool->mem_pool)->reclaimable += b->buflen;
+                }
+                /* unlink from current list (should call unlink_buffer?) */
+                if (b->prev || b->next) {
+                    b->prev->next = b->next;
+                    b->next->prev = b->prev;
                 }
                 add_free_buffer(interpreter, pool, b);
             } else if (!(b->flags & BUFFER_on_free_list_FLAG)) {
Index: hash.c
===================================================================
RCS file: /cvs/public/parrot/hash.c,v
retrieving revision 1.21
diff -u -r1.21 hash.c
--- hash.c	17 Aug 2002 01:11:08 -0000	1.21
+++ hash.c	17 Aug 2002 07:57:16 -0000
@@ -408,8 +408,9 @@
             /* FIXME: If string_compare triggers a collection, both
              * bucket and prev will end up pointing to junk memory.
              * They need to be BucketIndexes or refetched. */
-            if (prev)
+            if (prev) {
                 prev->next = bucket->next;
+            }
             else {
                 BucketIndex* table = (BucketIndex*) hash->buffer.bufstart;
                 table[slot] = bucket->next;
Index: headers.c
===================================================================
RCS file: /cvs/public/parrot/headers.c,v
retrieving revision 1.7
diff -u -r1.7 headers.c
--- headers.c	12 Aug 2002 06:55:02 -0000	1.7
+++ headers.c	17 Aug 2002 07:57:30 -0000
@@ -56,6 +56,7 @@
     pmc->flags = 0;
     /* Make sure it doesn't seem to be on the GC list */
     pmc->next_for_GC = NULL;
+    pmc->last_seen = interpreter->cycle;
     
     return pmc;
 }
@@ -101,6 +102,9 @@
     /* Don't let it point to garbage memory */
     buffer->bufstart = NULL;
     buffer->flags = 0;
+    buffer->pool = pool;
+    buffer->next = buffer->prev = NULL;
+    buffer->last_seen = interpreter->cycle;
     
     return buffer;
 }
@@ -272,6 +276,27 @@
     return buffer;
 }
 
+/* Create a new buffer header sharing data with an existing buffer 
+ * Unchecked precondition: offset+length must be within the source buffer */
+Buffer *
+new_shared_header(struct Parrot_Interp *interpreter, Buffer *src, 
+                  UINTVAL offset, UINTVAL length)
+{
+    Buffer *b;
+
+    /* get buffer header from the free pool */
+    b = get_free_buffer(interpreter, src->pool);
+    /* Point to the right place */
+    b->buflen = length;
+    b->bufstart = (char *)src->bufstart + offset;
+    /* Link it in bufstart sequence */
+    b->prev = src;
+    b->next = src->next;
+    b->next->prev = b;
+    b->prev->next = b;
+    /* Return it */
+    return b;
+}
 
 size_t
 get_max_buffer_address(struct Parrot_Interp *interpreter)
Index: interpreter.c
===================================================================
RCS file: /cvs/public/parrot/interpreter.c,v
retrieving revision 1.96
diff -u -r1.96 interpreter.c
--- interpreter.c	17 Aug 2002 01:11:08 -0000	1.96
+++ interpreter.c	17 Aug 2002 07:57:43 -0000
@@ -504,6 +504,7 @@
     interpreter->memory_collected = 0;
     interpreter->DOD_block_level = 1;
     interpreter->GC_block_level = 1;
+    interpreter->cycle = 0;
 
     /* Set up the memory allocation system */
     mem_setup_allocator(interpreter);
Index: resources.c
===================================================================
RCS file: /cvs/public/parrot/resources.c,v
retrieving revision 1.78
diff -u -r1.78 resources.c
--- resources.c	12 Aug 2002 06:55:02 -0000	1.78
+++ resources.c	17 Aug 2002 07:57:46 -0000
@@ -54,6 +54,10 @@
     new_block->start = (char *)new_block + sizeof(struct Memory_Block);
     new_block->top = new_block->start;
 
+    /* Initialise the linked list of buffer headers */
+    new_block->header_list.next = new_block->header_list.prev = 
+        (Buffer *)&new_block->header_list;
+
     /* Note that we've allocated it */
     interpreter->memory_allocated += alloc_size;
 
@@ -70,9 +74,28 @@
     return new_block;
 }
 
+/* Link buffer to specified list */
+inline static void
+link_buffer(Buffer *buffer, struct Buffer_List *list)
+{
+    buffer->prev = list->prev;
+    list->prev->next = buffer;
+    buffer->next = (Buffer *)list;
+    list->prev = buffer;
+}
+
+/* Unlink buffer from current list */
+inline static void
+unlink_buffer(Buffer *buffer)
+{
+    buffer->prev->next = buffer->next;
+    buffer->next->prev = buffer->prev;
+}
+
+
 static void *
-mem_allocate(struct Parrot_Interp *interpreter, size_t *req_size,
-             struct Memory_Pool *pool, size_t align_1)
+mem_allocate(struct Parrot_Interp *interpreter, Buffer *buffer, 
+             size_t *req_size, struct Memory_Pool *pool, size_t align_1)
 {
     char *return_val;
     size_t size = *req_size;
@@ -125,11 +148,17 @@
     pool->top_block->top += size;
     pool->top_block->free -= size;
     *req_size = size;
+
+    /* Unlink buffer from previous list */
+    if (buffer->prev) unlink_buffer(buffer);
+
+    /* Link buffer to current memory page */
+    link_buffer(buffer, &pool->top_block->header_list);
+
     return (void *)return_val;
 }
 
 
-
 /** Compaction Code **/
 
 /* Compact the buffer pool */
@@ -140,10 +169,17 @@
     struct Memory_Block *new_block;        /* A pointer to our working block */
     char *cur_spot;               /* Where we're currently copying to */
     UINTVAL cur_size;     /* How big our chunk is going to be */
-    struct Small_Object_Arena *cur_buffer_arena;
-    struct Small_Object_Pool *header_pool;
-    INTVAL j;
-    UINTVAL object_size;
+    struct Memory_Block *page;
+    struct Memory_Block *next_page;
+    Buffer *b;
+    Buffer *next_buffer;
+    struct Memory_Block *old_top;   /* top block before collection */
+    Buffer* last_header;
+    char* last_start;
+    char* last_end;
+    char* new_start;
+    char* new_end;
+
     /* Bail if we're blocked */
     if (interpreter->GC_block_level) {
         return;
@@ -155,11 +191,16 @@
     interpreter->header_allocs_since_last_collect = 0;
     interpreter->collect_runs++;
 
+    /* Store the current top of pool */
+    old_top = pool->top_block;
+
     /* total-reclaimable == currently used. Add a minimum block to the
      * current amount, so we can avoid having to allocate it in the
      * future. */
+/*
     total_size = pool->total_allocated - pool->reclaimable + pool->minimum_block_size;
-    /* total_size = pool->total_allocated; */
+*/
+     total_size = pool->total_allocated; 
     /* TODO: can reduce this by pool->total_reclaimable if we want to 
      * be precise */
     /* Snag a block big enough for everything */
@@ -168,54 +209,53 @@
     /* Start at the beginning */
     cur_spot = new_block->start;
   
-    /* Run through all the Buffer header pools and copy */
-    for (j = -2; j < (INTVAL) interpreter->arena_base->num_sized; j++) {
-        if (j == -2) header_pool = interpreter->arena_base->buffer_header_pool;
-        else if (j == -1) header_pool = interpreter->arena_base->string_header_pool;
-        else header_pool = interpreter->arena_base->sized_header_pools[j];
-        if (header_pool == NULL) continue;
-        
-        object_size = header_pool->object_size;
-        
-        for (cur_buffer_arena = header_pool->last_Arena;
-             NULL != cur_buffer_arena;
-             cur_buffer_arena = cur_buffer_arena->prev)
+    /* Run through all the Buffer pools and copy */
+    for (page = old_top; page; page = next_page) 
+    {
+        next_page = page->prev;
+        last_start = NULL;
+        last_end = NULL;
+        last_header = NULL;
+        for (b = page->header_list.next; b != (Buffer *)&page->header_list; 
+             b = next_buffer) 
         {
-            Buffer *b = cur_buffer_arena->start_objects;
-            UINTVAL i;
-            for (i = 0; i < cur_buffer_arena->used; i++) {
-                if (b->bufstart) {
-                    /* Is the buffer live, and can we move it? */
-                    if (!(b->flags & (BUFFER_on_free_list_FLAG | 
-                                      BUFFER_constant_FLAG | 
-                                      BUFFER_immobile_FLAG))) 
-                    {
-                        memcpy(cur_spot, b->bufstart, b->buflen);
-                        b->bufstart = cur_spot;
-                        cur_size = b->buflen;
-                        cur_size = (cur_size + header_pool->align_1) & ~header_pool->align_1;
-                        cur_spot += cur_size;
-                    }
-                }
-                b = (Buffer *)((char *)b + object_size);
-            }
-        }
-    }
-
-    /* Run through all the out-of-band Buffer header pools and copy */
-    for (j = 0; j < (INTVAL)( interpreter->arena_base->extra_buffer_headers.buflen / sizeof(Buffer*) ); j++) {
-        Buffer** buffers = interpreter->arena_base->extra_buffer_headers.bufstart;
-        Buffer* b = buffers[j];
-        if (b->bufstart) {
+            /* Get link to next buffer before we move the current one */
+            next_buffer = b->next;
+            /* Is the buffer live, and can we move it? */
             if (!(b->flags & (BUFFER_on_free_list_FLAG | 
                               BUFFER_constant_FLAG | 
-                              BUFFER_immobile_FLAG)))
+                              BUFFER_immobile_FLAG))) 
             {
-                memcpy(cur_spot, b->bufstart, b->buflen);
-                b->bufstart = cur_spot;
-                cur_size = b->buflen;
-                cur_size = (cur_size + BUFFER_ALIGNMENT - 1) & ~(BUFFER_ALIGNMENT - 1);
-                cur_spot += cur_size;
+                /* Keep sharing if possible */
+                if ((char *)b->bufstart >= last_start && 
+                    (char *)b->bufstart + b->buflen <= last_end) {
+                    size_t offset = (char *)b->bufstart - last_start;
+                    b->bufstart = new_start + offset;
+                    b->flags |= BUFFER_COW_FLAG;
+                    last_header->flags |= BUFFER_COW_FLAG;
+                    unlink_buffer(b);
+                    link_buffer(b, &pool->top_block->header_list);
+                }
+                else {
+                    last_start = b->bufstart;
+                    last_end = last_start + b->buflen;
+                    last_header = b;
+
+                    new_start = cur_spot;
+                    cur_size = b->buflen;
+                    if (b->pool)
+                        cur_size = (cur_size + b->pool->align_1) & ~b->pool->align_1;
+                    else
+                        cur_size = (cur_size + 3) & ~3;
+                    cur_spot += cur_size;
+                    unlink_buffer(b);
+                    link_buffer(b, &pool->top_block->header_list);
+
+                    new_end = new_start + b->buflen;
+                    memcpy(new_start, b->bufstart, b->buflen);
+                    b->bufstart = new_start;
+                    b->flags &= ~(UINTVAL)BUFFER_COW_FLAG;
+                }
             }
         }
     }
@@ -284,11 +324,11 @@
     if (interpreter) {
         interpreter->arena_base->memory_pool->reclaimable +=
             buffer->buflen;
-        mem = mem_allocate(interpreter, &alloc_size, 
+        mem = mem_allocate(interpreter, buffer, &alloc_size, 
                            interpreter->arena_base->memory_pool, BUFFER_ALIGNMENT-1);
     }
     else {
-        mem = mem_allocate(NULL, &alloc_size, NULL, BUFFER_ALIGNMENT-1);
+        mem = mem_allocate(NULL, buffer, &alloc_size, NULL, BUFFER_ALIGNMENT-1);
     }
 
     if (!mem) {
@@ -322,7 +362,8 @@
          : interpreter->arena_base->memory_pool;
     pool->reclaimable += str->buflen;
 
-    mem = mem_allocate(interpreter, &alloc_size, pool, STRING_ALIGNMENT-1);
+    mem = mem_allocate(interpreter, (Buffer *)str, &alloc_size, pool, 
+                       STRING_ALIGNMENT-1);
     if (!mem) {
         return NULL;
     }
@@ -344,7 +385,7 @@
     size_t req_size = size;
     ((Buffer *)buffer)->buflen = 0;
     ((Buffer *)buffer)->bufstart = NULL;
-    ((Buffer *)buffer)->bufstart = mem_allocate(interpreter, &req_size, 
+    ((Buffer *)buffer)->bufstart = mem_allocate(interpreter, buffer, &req_size, 
                        interpreter->arena_base->memory_pool, BUFFER_ALIGNMENT-1);
     ((Buffer *)buffer)->buflen = size;
     return buffer;
@@ -363,13 +404,15 @@
     str->bufstart = NULL;
 
     if (!interpreter) {
-        str->bufstart = mem_allocate(NULL, &req_size, NULL, STRING_ALIGNMENT-1);
+        str->bufstart = mem_allocate(NULL, (Buffer *)str, &req_size, NULL, 
+                                     STRING_ALIGNMENT-1);
     }
     else {
         pool = (str->flags & BUFFER_constant_FLAG)
              ? interpreter->arena_base->constant_string_pool
              : interpreter->arena_base->memory_pool;
-        str->bufstart = mem_allocate(interpreter, &req_size, pool, STRING_ALIGNMENT-1);
+        str->bufstart = mem_allocate(interpreter, (Buffer *)str, &req_size, 
+                                     pool, STRING_ALIGNMENT-1);
     }
     str->buflen = req_size;
     return str;
Index: string.c
===================================================================
RCS file: /cvs/public/parrot/string.c,v
retrieving revision 1.83
diff -u -r1.83 string.c
--- string.c	23 Jul 2002 02:09:27 -0000	1.83
+++ string.c	17 Aug 2002 07:58:24 -0000
@@ -29,6 +29,17 @@
     string_unicode_type = chartype_lookup("unicode");
 }
 
+inline static UINTVAL
+free_space(const STRING *s)
+{
+    if (s->flags & BUFFER_constant_FLAG || s->flags & BUFFER_COW_FLAG) {
+        return 0;
+    }
+    else {
+        return s->buflen - s->bufused;
+    }
+}
+
 /*=for api string string_append
  * Take in two strings and append the second string to the first
  */
@@ -58,7 +69,7 @@
                                  NULL);
         }
         /* make sure A's big enough for both */
-        if (a->buflen < a->bufused + b->bufused) {
+        if (free_space(a) < b->bufused) {
             a = string_grow(interpreter, a, ((a->bufused + b->bufused)
                             - a->buflen) + EXTRA_SIZE);
         }
@@ -219,20 +230,29 @@
  * create a copy of the argument passed in
  */
 STRING *
-string_copy(struct Parrot_Interp *interpreter, const STRING *s)
+string_copy(struct Parrot_Interp *interpreter, STRING *s)
 {
     STRING *d;
-    d = new_string_header(interpreter, 
-                          s->flags & ~(UINTVAL)BUFFER_constant_FLAG);
-    Parrot_allocate_string(interpreter, d, s->buflen);
+
+    if (s->flags & BUFFER_constant_FLAG || !s->pool) {
+        d = new_string_header(interpreter, 
+                              s->flags & ~(UINTVAL)BUFFER_constant_FLAG);
+        Parrot_allocate_string(interpreter, d, s->buflen);
+        memcpy(d->bufstart, s->bufstart, s->buflen);
+    }
+    else {
+        d = (STRING *)new_shared_header(interpreter, (Buffer *)s, 0, 
+                                        s->bufused);
+        s->flags |= BUFFER_COW_FLAG;
+        d->flags = s->flags & ~(UINTVAL)BUFFER_constant_FLAG;
+    }
+
     d->bufused = s->bufused;
     d->strlen = s->strlen;
     d->encoding = s->encoding;
     d->type = s->type;
     d->language = s->language;
 
-    memcpy(d->bufstart, s->bufstart, s->buflen);
-
     return d;
 }
 
@@ -407,7 +427,7 @@
  * Allocate memory for d if necessary.
  */
 STRING *
-string_substr(struct Parrot_Interp *interpreter, const STRING *src,
+string_substr(struct Parrot_Interp *interpreter, STRING *src,
               INTVAL offset, INTVAL length, STRING **d)
 {
     STRING *dest;
@@ -447,17 +467,27 @@
                                             true_length) -
         (char *)src->bufstart;
 
-    dest =
-        string_make(interpreter, NULL, true_length * src->encoding->max_bytes,
-                    src->encoding, 0, src->type);
-
     if (subend_off < substart_off) {
         internal_exception(SUBSTR_OUT_OF_STRING,
                            "subend somehow is less than substart");
     }
 
-    mem_sys_memcopy(dest->bufstart, (char *)src->bufstart + substart_off,
-                    (unsigned)(subend_off - substart_off));
+    if (src->flags & BUFFER_constant_FLAG) {
+        dest = string_make(interpreter, NULL, 
+                           true_length * src->encoding->max_bytes,
+                           src->encoding, 0, src->type);
+        mem_sys_memcopy(dest->bufstart, (char *)src->bufstart + substart_off,
+                        (unsigned)(subend_off - substart_off));
+    }
+    else {
+        dest = (STRING *)new_shared_header(interpreter, (Buffer *)src, 
+                                           substart_off, true_length);
+        src->flags |= BUFFER_COW_FLAG;
+        dest->flags = src->flags & ~(UINTVAL)BUFFER_constant_FLAG;
+        dest->encoding = src->encoding;
+        dest->type = src->type;
+        dest->language = src->language;
+    }
     dest->bufused = subend_off - substart_off;
     dest->strlen = true_length;
 
@@ -489,6 +519,11 @@
     UINTVAL true_offset;
     UINTVAL true_length;
     INTVAL diff;
+
+    /* TODO do this only when required */
+    if (src->flags & BUFFER_COW_FLAG) {
+        src = string_grow(interpreter, src, 0);
+    }
         
     true_offset = (UINTVAL)offset;
     true_length = (UINTVAL)length;
@@ -880,7 +915,7 @@
 {
     char *cstring;
 
-    if (s->buflen == s->bufused)
+    if (s->buflen == s->bufused || s->flags & BUFFER_COW_FLAG) 
         string_grow(interpreter, s, 1);
 
     cstring = s->bufstart;
Index: include/parrot/headers.h
===================================================================
RCS file: /cvs/public/parrot/include/parrot/headers.h,v
retrieving revision 1.5
diff -u -r1.5 headers.h
--- include/parrot/headers.h	23 Jul 2002 07:26:09 -0000	1.5
+++ include/parrot/headers.h	17 Aug 2002 07:58:37 -0000
@@ -63,6 +63,8 @@
 STRING *new_string_header(struct Parrot_Interp *interpreter, UINTVAL flags);
 Buffer *new_buffer_header(struct Parrot_Interp *interpreter);
 void *new_bufferlike_header(struct Parrot_Interp *interpreter, size_t size);
+Buffer *new_shared_header(struct Parrot_Interp *interpreter, Buffer *src, 
+                          UINTVAL offset, UINTVAL length);
 
 size_t get_max_buffer_address(struct Parrot_Interp *interpreter);
 size_t get_min_buffer_address(struct Parrot_Interp *interpreter);
Index: include/parrot/interpreter.h
===================================================================
RCS file: /cvs/public/parrot/include/parrot/interpreter.h,v
retrieving revision 1.52
diff -u -r1.52 interpreter.h
--- include/parrot/interpreter.h	3 Aug 2002 07:35:36 -0000	1.52
+++ include/parrot/interpreter.h	17 Aug 2002 07:58:44 -0000
@@ -173,6 +173,7 @@
                                    requests are there? */
     UINTVAL GC_block_level;     /* How many outstanding GC block
                                    requests are there? */
+    UINTVAL cycle;              /* opcode counter for GC system */
 
     PDB_t *pdb;                 /* Debug system */
     void *lo_var_ptr;           /* Pointer to memory on runops system stack */
Index: include/parrot/pmc.h
===================================================================
RCS file: /cvs/public/parrot/include/parrot/pmc.h,v
retrieving revision 1.36
diff -u -r1.36 pmc.h
--- include/parrot/pmc.h	15 Aug 2002 06:44:09 -0000	1.36
+++ include/parrot/pmc.h	17 Aug 2002 07:58:45 -0000
@@ -65,6 +65,7 @@
                                  dereference when setting, but let us
                                  memset the actual GC data in a big
                                  block */
+    UINTVAL last_seen;
 };
 
 /* PMC flag bits */
Index: include/parrot/resources.h
===================================================================
RCS file: /cvs/public/parrot/include/parrot/resources.h,v
retrieving revision 1.37
diff -u -r1.37 resources.h
--- include/parrot/resources.h	3 Aug 2002 07:35:36 -0000	1.37
+++ include/parrot/resources.h	17 Aug 2002 07:58:45 -0000
@@ -15,6 +15,12 @@
 
 #include "parrot/parrot.h"
 
+/* Must be the same fields as at the start of a buffer */
+struct Buffer_List {
+    Buffer *next;
+    Buffer *prev;
+};
+
 struct Memory_Block {
     size_t free;
     size_t size;
@@ -22,6 +28,7 @@
     struct Memory_Block *next;
     char *start;
     char *top;
+    struct Buffer_List header_list;
 };
 
 struct Memory_Pool {
Index: include/parrot/string.h
===================================================================
RCS file: /cvs/public/parrot/include/parrot/string.h,v
retrieving revision 1.43
diff -u -r1.43 string.h
--- include/parrot/string.h	18 Jul 2002 04:30:42 -0000	1.43
+++ include/parrot/string.h	17 Aug 2002 07:58:53 -0000
@@ -22,9 +22,13 @@
 #ifdef PARROT_IN_CORE
 
 struct parrot_string_t {
+    struct parrot_string_t *next;
+    struct parrot_string_t *prev;
+    struct Small_Object_Pool *pool;
     void *bufstart;
     UINTVAL buflen;
     UINTVAL flags;
+    UINTVAL last_seen;
     UINTVAL bufused;
     UINTVAL strlen;
     const ENCODING *encoding;
@@ -37,10 +41,14 @@
 #include "parrot/parrot.h"
 
 /* Generic buffer header. Enough to GC */
-typedef struct {
+typedef struct parrot_buffer_t {
+    struct parrot_buffer_t *next;
+    struct parrot_buffer_t *prev;
+    struct Small_Object_Pool *pool;
     void *bufstart;
     UINTVAL buflen;
     UINTVAL flags;
+    UINTVAL last_seen;
 } Buffer;
 
 typedef struct parrot_string_t String;
Index: include/parrot/string_funcs.h
===================================================================
RCS file: /cvs/public/parrot/include/parrot/string_funcs.h,v
retrieving revision 1.13
diff -u -r1.13 string_funcs.h
--- include/parrot/string_funcs.h	24 Jun 2002 16:41:28 -0000	1.13
+++ include/parrot/string_funcs.h	17 Aug 2002 07:58:53 -0000
@@ -23,7 +23,7 @@
 STRING *string_repeat(struct Parrot_Interp *, const STRING *, UINTVAL,
                       STRING **);
 STRING *string_chopn(STRING *, INTVAL);
-STRING *string_substr(struct Parrot_Interp *, const STRING *, INTVAL,
+STRING *string_substr(struct Parrot_Interp *, STRING *, INTVAL,
                       INTVAL, STRING **);
 STRING *string_replace(struct Parrot_Interp *, STRING *, INTVAL, INTVAL,
                        STRING *, STRING **);
@@ -43,7 +43,7 @@
 STRING *string_make(struct Parrot_Interp *, const void *buffer,
                     UINTVAL buflen, const ENCODING *, UINTVAL flags,
                     const CHARTYPE *);
-STRING *string_copy(struct Parrot_Interp *, const STRING *);
+STRING *string_copy(struct Parrot_Interp *, STRING *);
 STRING *string_transcode(struct Parrot_Interp *, const STRING *src,
                          const ENCODING *, const CHARTYPE *,
                          STRING **dest_ptr);
Index: lib/Parrot/OpsFile.pm
===================================================================
RCS file: /cvs/public/parrot/lib/Parrot/OpsFile.pm,v
retrieving revision 1.25
diff -u -r1.25 OpsFile.pm
--- lib/Parrot/OpsFile.pm	30 Jun 2002 14:21:05 -0000	1.25
+++ lib/Parrot/OpsFile.pm	17 Aug 2002 07:59:07 -0000
@@ -262,6 +262,14 @@
   my $next = 0;
   my $jumps = 0;
 
+  # The following code alters the body of opcodes that appear to contain
+  # functions calls to increment an interpreter-level counter. This counter 
+  # is used by the GC system to avoid reaping objects created during 
+  # execution of the current op.
+  if ($body =~ m/[0-z]\(+[0-z]/) {
+    $body = "  interpreter->cycle++;\n{\n" . $body . "}";
+  }
+
   foreach my $variant (expand_args(@$args)) {
       my(@fixedargs)=split(/,/,$variant);
       my $op = Parrot::Op->new($code++, $type, $short_name,
